import contextvars
import csv
import datetime
import io
import os
from abc import ABC, abstractmethod
from typing import Annotated, AsyncGenerator, Literal, Optional
from uuid import UUID
from pydantic import validate_call
import aiofiles
import httpx
import jwt
from loguru import logger

from ._models import Dataset, Field, create_model, SchemaProperty, DatasetSchema

from ._protcols import SyncInstallationContextProtocol, AsyncInstallationContextProtocol


class DatasetManager:
    def __init__(self, installation_context: SyncInstallationContextProtocol):
        self.installation_context = installation_context
        self.client = installation_context.client
        self.session = installation_context.session
        self.datasets = None

    def __iter__(self):
        response = self.session.get(self.client.base_url.format("datasets"))
        response.raise_for_status()
        data = response.json()
        return iter([Dataset(**item) for item in data.get("data", [])])

    def find(self, name: str):
        for dataset in self:
            if dataset.name == name:
                return DatasetOperations(dataset, self.installation_context)
        raise KeyError(f"Dataset with name '{name}' not found")

    @validate_call(validate_return=True)
    def get(self, dataset_id: str) -> Dataset:
        response = self.session.get(
            self.client.base_url.format(f"datasets/{dataset_id}")
        )
        response.raise_for_status()
        return response.json()

    def create(
        self,
        name: str,
        description: str,
        schema: dict | DatasetSchema,
    ) -> "DatasetOperations":
        if isinstance(schema, dict):
            schema = DatasetSchema(**schema)

        response = self.session.post(
            self.client.base_url.format("datasets"),
            json={
                "name": name,
                "description": description,
                "schema": schema.model_dump(
                    mode="json",
                    exclude_none=True,
                    by_alias=True,
                ),
            },
        )
        response.raise_for_status()
        data = response.json()
        dataset = Dataset(**data)
        if self.datasets is not None:
            self.datasets.append(dataset)
        return DatasetOperations(dataset, self.installation_context)


class AsyncDatasetManager:
    def __init__(self, installation_context: AsyncInstallationContextProtocol):
        self.installation_context = installation_context
        self.client = installation_context.client
        self.session = installation_context.session
        self.datasets = None

    async def __aiter__(self) -> AsyncGenerator["AsyncDatasetOperations", None]:
        async for dataset in self.list_datasets():
            yield AsyncDatasetOperations(dataset, self.installation_context)

    async def list_datasets(self) -> AsyncGenerator[Dataset, None]:
        response = await self.session.get(self.client.base_url.format("datasets"))
        response.raise_for_status()
        data = response.json()
        for item in data.get("data", []):
            yield Dataset(**item)
        # return [Dataset(**item) for item in data.get("data", [])]

    async def find(self, name: str) -> "AsyncDatasetOperations":
        async for dataset in self:
            if dataset.name == name:
                return AsyncDatasetOperations(dataset, self.installation_context)
        raise KeyError(f"Dataset with name '{name}' not found")

    @validate_call(validate_return=True)
    async def get(self, dataset_id: str) -> Dataset:
        response = await self.session.get(
            self.client.base_url.format(f"datasets/{dataset_id}")
        )
        response.raise_for_status()
        return response.json()

    async def create(
        self,
        name: str,
        description: str,
        schema: dict | DatasetSchema,
    ) -> "DatasetOperations":
        if isinstance(schema, dict):
            schema = DatasetSchema(**schema)
        response = await self.session.post(
            self.client.base_url.format("datasets"),
            json={
                "name": name,
                "description": description,
                "schema": schema.model_dump(
                    mode="json", exclude_none=True, by_alias=True
                ),
            },
        )
        response.raise_for_status()
        data = response.json()
        dataset = Dataset(**data)
        if self.datasets is not None:
            self.datasets.append(dataset)
        return AsyncDatasetOperations(dataset, self.installation_context)


class ABCDatasetOperations(ABC):
    _dataset: Dataset

    @property
    def dataset_id(self) -> str:
        return self._dataset.dataset_id

    @property
    def name(self) -> str:
        return self._dataset.name

    @property
    def description(self):
        return self._dataset.description

    @property
    def date_created(self):
        return self._dataset.date_created

    @property
    def record_count(self):
        return self._dataset.record_count

    @property
    def created_by(self):
        return self._dataset.created_by

    @property
    def created_by_workspace(self):
        return self._dataset.created_by_workspace

    @property
    def dataset_tags(self):
        return self._dataset.dataset_tags

    @property
    def dataset_schema(self):
        return self._dataset.dataset_schema

    def _create_record_model(self):
        _schema = []
        for prop in self.dataset_schema.properties:
            _schema.append((prop.name, (str, Field(default=None))))
        return create_model("Record", **dict(_schema))

    @abstractmethod
    def records(self) -> list[dict]:
        raise NotImplementedError

    @abstractmethod
    def load(self, data: list[dict]):
        raise NotImplementedError

    @abstractmethod
    def load_from_file(self, file_path: str):
        raise NotImplementedError

    @abstractmethod
    def load_from_url(self, data_url: str):
        raise NotImplementedError

    @abstractmethod
    def get_upload_url(self):
        raise NotImplementedError

    @abstractmethod
    def upload_data_to_url(self, upload_url: str, data: list[dict]):
        raise NotImplementedError

    @abstractmethod
    def upload_file_to_url(self, upload_url: str, file_path: str):
        raise NotImplementedError

    @abstractmethod
    def get_download_url(self):
        raise NotImplementedError

    @abstractmethod
    def download_data_from_url(self, download_url: str) -> list[dict]:
        raise NotImplementedError

    def __repr__(self):
        return f"Operations[Dataset(dataset_id={self.dataset_id}, name={self.name})]"


class DatasetOperations(ABCDatasetOperations):
    def __init__(
        self, dataset: Dataset, installation_context: SyncInstallationContextProtocol
    ):
        self._dataset = dataset
        self._install_ctx = installation_context

    @property
    def client(self):
        return self._install_ctx.client

    @property
    def session(self):
        return self.client.session

    def load(self, data: list[dict], validate_records: bool = False):
        try:
            _model = self._create_record_model()
            if validate_records:
                _data = [_model(**record).model_dump(mode="json") for record in data]
            else:
                _data = data
        except Exception as e:
            raise ValueError(f"Data does not match schema: {e}")

        upload_info = self.get_upload_url()
        upload_url = upload_info["url"]
        r = self.upload_data_to_url(upload_url, _data)
        assert r.status_code == 200

    def get_upload_url(self):
        response = self.session.post(
            self.client.base_url.format(f"datasets/{self.dataset_id}/uploadUrl"),
            json={"contentType": "csv"},
        )
        response.raise_for_status()
        return response.json()

    def upload_data_to_url(self, upload_url: str, data: list[dict]):
        _buffer = io.StringIO()
        writer = csv.DictWriter(_buffer, fieldnames=data[0].keys())
        writer.writeheader()
        writer.writerows(data)
        _buffer.seek(0)
        response = self.session.put(
            upload_url, data=_buffer.getvalue(), headers={"Content-Type": "text/csv"}
        )
        response.raise_for_status()
        return response

    def records(self):
        download_info = self.get_download_url()
        download_url = download_info["url"]
        return self.download_data_from_url(download_url)

    def get_download_url(self):
        response = self.session.post(
            self.client.base_url.format(f"datasets/{self.dataset_id}/records:retrieve"),
            json={},
        )
        response.raise_for_status()
        return response.json()

    def download_data_from_url(self, download_url: str) -> list[dict]:
        response = self.session.get(download_url)
        response.raise_for_status()
        with io.StringIO(response.text) as f:
            return list(csv.DictReader(f))

    def load_from_file(self, file_path: str):
        upload_info = self.get_upload_url_sync()
        upload_url = upload_info["url"]
        self.upload_file_to_url_sync(upload_url, file_path)

    def upload_file_to_url(self, upload_url: str, file_path: str):
        with open(file_path, "rb") as f:
            response = self.session.put(upload_url, data=f)
        response.raise_for_status()

    def load_from_url(self, data_url: str):
        response = self.session.post(
            self.client.base_url.format(
                f"datasets/{self.dataset_id}/records:load?mode=replace"
            ),
            json={"url": data_url},
        )
        response.raise_for_status()
        return response.json()


class AsyncDatasetOperations(ABCDatasetOperations):
    def __init__(
        self,
        dataset: Dataset,
        installation_context: AsyncInstallationContextProtocol,
    ):
        self._dataset = dataset
        self._install_ctx = installation_context

    @property
    def client(self):
        return self._install_ctx.client

    @property
    def session(self):
        return self.client.asession

    async def load(self, data: list[dict], validate_records: bool = True):
        try:
            _model = self._create_record_model()
            if validate_records:
                _data = [_model(**record).model_dump(mode="json") for record in data]
            else:
                _data = data
        except Exception as e:
            raise ValueError(f"Data does not match schema: {e}")

        upload_info = await self.get_upload_url()
        upload_url = upload_info["url"]
        r = await self.upload_data_to_url(upload_url, _data)
        assert r.status_code == 200
        return True

    async def get_upload_url(self):
        response = await self.session.post(
            self.client.base_url.format(f"datasets/{self.dataset_id}/uploadUrl"),
            json={"contentType": "csv"},
        )
        response.raise_for_status()
        return response.json()

    async def upload_data_to_url(self, upload_url: str, data: list[dict]):
        _buffer = io.StringIO()
        writer = csv.DictWriter(_buffer, fieldnames=data[0].keys())
        writer.writeheader()
        writer.writerows(data)
        _buffer.seek(0)
        response = await self.session.put(
            upload_url, data=_buffer.getvalue(), headers={"Content-Type": "text/csv"}
        )
        response.raise_for_status()
        return response

    async def load_data_from_public_url(self, url: str):
        response = await self.session.post(
            self.client.base_url.format(
                f"datasets/{self.dataset_id}/records:load?mode=replace"
            ),
            json={"url": url},
        )
        response.raise_for_status()
        return response.json()

    async def records(self) -> AsyncGenerator[dict, None]:
        download_info = await self.get_download_url()
        logger.info(download_info)
        download_url = download_info["url"]
        logger.info(download_url)
        r = await self.download_data_from_url(download_url)
        for record in r:
            yield record

    async def get_download_url(self):
        response = await self.session.post(
            self.client.base_url.format(f"datasets/{self.dataset_id}/records:retrieve"),
            json={"contentType": "csv"},
        )
        response.raise_for_status()
        return response.json()

    async def download_data_from_url(self, download_url: str) -> list[dict]:
        response = await self.session.get(download_url)
        response.raise_for_status()
        with io.StringIO(response.text) as f:
            return list(csv.DictReader(f))

    async def load_from_file(self, file_path: str):
        upload_info = await self.get_upload_url()
        upload_url = upload_info["url"]
        await self.upload_file_to_url(upload_url, file_path)

    async def upload_file_to_url(self, upload_url: str, file_path: str):
        async with aiofiles.open(file_path, "rb") as f:
            file_data = await f.read()
            response = await self.session.put(upload_url, data=file_data)
        response.raise_for_status()

    async def load_from_url(self, data_url: str):
        response = await self.session.post(
            self.client.base_url.format(
                f"datasets/{self.dataset_id}/records:load?mode=replace"
            ),
            json={"url": data_url},
        )
        response.raise_for_status()
        return response.json()
