from typing import Annotated, AsyncGenerator, Literal, Optional, Generic, TypeVar
import datetime
from uuid import UUID
from pydantic import BaseModel, Field, create_model, validate_call, ConfigDict


class _BaseModel(BaseModel):
    model_config = ConfigDict(
        populate_by_name=True,
    )


class WhoAmI(_BaseModel):
    model_config = ConfigDict(
        populate_by_name=True,
    )

    user_id: Annotated[int, Field(alias="userId")]
    workspace_id: Annotated[int, Field(alias="workspaceId")]
    app_id: Annotated[UUID, Field(alias="appId")]
    user_name: Annotated[str | None, Field(alias="userName")] = None
    email: Annotated[str | None, Field(alias="email")] = None
    date_terms_accepted: Annotated[
        datetime.datetime | None, Field(alias="dateTermsAccepted")
    ] = None


class CreatedBy(_BaseModel):
    user_id: Annotated[int, Field(alias="userId")]
    display_name: Annotated[str, Field(alias="displayName")]
    email: Annotated[str | None, Field(default=None)] = None


class Workspace(_BaseModel):
    workspace_id: Annotated[int, Field(alias="workspaceId")]
    display_name: Annotated[str, Field(alias="displayName")]


class SchemaProperty(_BaseModel):
    type: Literal["string"] = "string"
    required: Annotated[bool, Field(exclude=True)]
    name: str


class DatasetSchema(_BaseModel):
    dataset_schema_id: Annotated[int | None, Field(alias="datasetSchemaId")] = None
    properties: list[SchemaProperty]
    primary_key: Annotated[
        list[str], Field(alias="primaryKey", default_factory=list)
    ] = []


class CreateDataset(_BaseModel):
    name: str = Field(
        ..., title="Name of the dataset", description="Name of the dataset"
    )
    description: str = Field(
        ...,
        title="Description of the dataset",
        description="Description of the dataset",
    )
    schema_: DatasetSchema = Field(
        ...,
        alias="schema",
        title="Schema of the dataset",
        description="Schema of the dataset",
    )


class Dataset(_BaseModel):
    dataset_id: Annotated[UUID, Field(alias="datasetId")]
    name: str
    description: str
    date_created: Annotated[datetime.datetime, Field(alias="dateCreated")]
    record_count: Annotated[int, Field(alias="recordCount")]
    created_by: Annotated[CreatedBy, Field(alias="createdBy")]
    created_by_workspace: Annotated[Workspace, Field(alias="createdByWorkspace")]
    dataset_tags: Annotated[list[str], Field(alias="datasetTags")] = []
    dataset_schema: Annotated[DatasetSchema, Field(alias="datasetSchema")]


class Installation(_BaseModel):
    movement_app_installation_id: Annotated[
        int, Field(alias="movementAppInstallationId")
    ]
    movement_app_id: Annotated[UUID, Field(alias="movementAppId")]
    workspace_id: Annotated[int, Field(alias="workspaceId")]
    created_by: Annotated[CreatedBy, Field(alias="createdBy")]
    date_created: Annotated[datetime.datetime, Field(alias="dateCreated")]
    name: str


class ClientToken(_BaseModel):
    token: str
    token_type: Annotated[Literal["Bearer"], Field(alias="tokenType")] = "Bearer"
    expires_at: Annotated[datetime.datetime, Field(alias="expiresAt")]


T = TypeVar("T", bound=BaseModel)


class ResponseMetadata(_BaseModel):
    count: int


class PaginatedResponse(_BaseModel, Generic[T]):
    metadata: ResponseMetadata
    data: list[T]


class ExceptionDetails(_BaseModel):
    details: str
    headers: dict[str, str] = Field(
        ...,
        title="Request headers",
        description="Request headers",
        default_factory=dict,
    )
    path: str
    endpoint: str
    routeValues: dict[str, str]


class ErrorResponse(_BaseModel):
    model_config = ConfigDict(extra="allow")

    type: str
    title: str
    status: int
    detail: str
    exception: ExceptionDetails
    source: str

    def __repr__(self):
        return f"<Error {self.title} ({self.status}): {self.detail} />"


class DXException(Exception):
    def __init__(self, error: ErrorResponse):
        self.error = error
        super().__init__()
